/**
 * Callback function for encrypting an unencrypted device.
 * It proceeds as follows:
 * 1.  check if the device isn't already encrypted;
 * 2.  mount the unencrypted device;
 * 3.  determine the total size of the files on the device, and guess if they
 *     will fit into memory;
 * 4.  rsync all files from the unencrypted device to memory;
 * 5.  check if an authorized_keys file can be found on the root device;
 * 6.  unmount the unencrypted device;
 * 7.  repartition the device:
 *     . an info partition to hold unencrypted data for use from the initrd,
 *       like the dropbear ssh host keys and user's ssh public keys;
 *     . a data partition for the actual encrypted root device;
 * 8.  create a filesystem on the info partition;
 * 9.  copy ssh host keys and the authorized_keys file from the initrd
 *     to the info partition;
 * 10. initialise a new encrypted container on the data partition, and unlock it
 *     for use;
 * 11. create a new filesystem on this new encrypted data partition,
 *     and mount it;
 * 12. rsync all files back from memory to the encrypted data partition;
 * 13. unmount the new filesystem, and lock the encrypted container;
 * 14. reboot.
 * @param[in]   request   incoming HTTP request
 * @param[out]  response  HTTP response to the request
 * @param[in]   user_data extra data to pass between handler and main thread
 * @return                internal status code
 */
int callback_encryption_init_post(const struct _u_request * request,
    struct _u_response * response, void * user_data)
{
    bool * reboot = (bool *)user_data;
    int r;

    // Read the encryption password from the request body.
    json_t * json_input = ulfius_get_json_body_request(request, NULL);
    const char * password = json_string_value(
        json_object_get(json_input, "password"));
    if (password == NULL)
    {
        json_decref(json_input);
        return send_simple_response(response, 400, "error", "missing password");
    }
 
    // Check if the device isn't already encrypted.
    // Actually we check if the device has any partitions; if so, we're not in
    // the expected situation of an unpartitioned unencrypted root device.
    if (path_exists(config.info_partition_device))
    {
        // The device is already encrypted; we don't want to encrypt it again.
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "already encrypted");
    }

    // Mount the filesystem on the unencrypted device.
    r = temporary_mount(config.root_device, UNENCRYPTED_TMP_MOUNTPOINT,
        FILESYSTEM_TYPE);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR, 
            "mounting root device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "mounting root device failed");
    }

    // Determine if the files on the unencrypted device would fit in memory.
    bool fits = filesystem_fits_in_memory(UNENCRYPTED_TMP_MOUNTPOINT,
        MEMORY_USAGE);
    if (! fits)
    {
        // Projected memory usage is really high, so abort.
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
           "device too large");
    }

    // Copy device contents to temporary filesystem.
    y_log_message(Y_LOG_LEVEL_DEBUG,
            "copying existing root device contents to memory");
    char * command = NULL;
    asprintf(&command, "rsync -a %s/ %s", UNENCRYPTED_TMP_MOUNTPOINT,
        config.tmp_data_location);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "copying root device contents into memory failed: return code %d",
            r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying root device contents into memory failed");
    }

    // Unmount unencrypted device.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting unencrypted device at %s",
        UNENCRYPTED_TMP_MOUNTPOINT);
    r = umount(UNENCRYPTED_TMP_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR, 
            "unmounting encrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unmounting unencrypted device failed");
    }

    // Check whether the authorized_keys file exists on root device.
    char * authorized_keys_path = NULL;
    asprintf(&authorized_keys_path, "%s%s",
        config.tmp_data_location, AUTHORIZED_KEYS_PATH);
    struct stat st = {0};
    if (stat(authorized_keys_path, &st) == -1)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "authorized_keys not found on root device at %s",
            authorized_keys_path);
        if (authorized_keys_path)
        {
            free(authorized_keys_path);
        }
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "authorized_keys not found on root device");
    }
    if (authorized_keys_path)
    {
        free(authorized_keys_path);
    }

    // Re-partition device.
    y_log_message(Y_LOG_LEVEL_DEBUG, "repartitioning device %s",
        config.root_device);
    command = NULL;
    asprintf(&command, "sgdisk -a 8192 -n 1:0:48M -N 2 %s", config.root_device);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "partitioning root device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "partitioning root device failed");
    }

    // Inform kernel of partitioning changes.
    y_log_message(Y_LOG_LEVEL_DEBUG, "running partprobe");
    command = NULL;
    asprintf(&command, "partprobe");
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR, "partprobe failed: return code %d", 
            r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "partprobe failed");
    }
    // Sleep for a little while.
    // This seems necessary for the newly created partitions to appear as
    // devices.
    y_log_message(Y_LOG_LEVEL_DEBUG, "waiting a bit...");
    sleep(2);

    // Create filesystem on the info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG, "creating filesystem on info partition");
    command = NULL;
    asprintf(&command, "mkfs -t %s %s", FILESYSTEM_TYPE,
        config.info_partition_device);
    y_log_message(Y_LOG_LEVEL_DEBUG, "command: %s", command);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "creating filesystem failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "creating filesystem inside encrypted container failed");
    }

    // Mount the info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG, "mounting info partition");
    r = temporary_mount(config.info_partition_device, INFO_TMP_MOUNTPOINT,
        FILESYSTEM_TYPE);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "mounting encrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "mounting encrypted root device failed");
    }

    // Create some directories in the info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG, "creating directories in info partition");
    command = NULL;
    asprintf(&command, "mkdir -p %s%s %s%s", INFO_TMP_MOUNTPOINT,
        AUTHORIZED_KEYS_DIR, INFO_TMP_MOUNTPOINT, SSH_HOST_KEY_DIR);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "creating directories failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "creating directories in info partition failed");
    }

    // Copy authorized_keys file to the info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG, "copying authorized_keys to info partition");
    command = NULL;
    asprintf(&command, "cp %s %s%s",
        AUTHORIZED_KEYS_PATH, INFO_TMP_MOUNTPOINT, AUTHORIZED_KEYS_PATH);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "copying authorized_keys failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying authorized_keys failed");
    }

    // Copy dropbear ssh host keys from the initrd to the info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG, "copying dropbear ssh host keys");
    command = NULL;
    asprintf(&command, "cp /etc/dropbear/* %s%s/", INFO_TMP_MOUNTPOINT,
        SSH_HOST_KEY_DIR);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_DEBUG,
            "copying dropbear ssh host keys failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying dropbear ssh host keys failed");
    }

    // Unmount info partition.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting info partition at %s", INFO_TMP_MOUNTPOINT);
    r = umount(INFO_TMP_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unmounting failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unmounting configuration partition failed");
    }

    // Initialise encrypted container on data partition.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "creating encrypted container at %s", config.data_partition_device);
    r = create_encrypted_device(config.data_partition_device, password);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "creating encrypted container failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "creating encryption container failed");
    }

    // Unlock the new container.
    y_log_message(Y_LOG_LEVEL_DEBUG, "unlocking encrypted device");
    r = encryption_unlock(config.data_partition_device,
        config.mapped_device_name, password);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unlocking encrypted container failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unlocking new encryption container failed");
    }

    // Create filesystem in the new container.
    y_log_message(Y_LOG_LEVEL_DEBUG, "creating filesystem in new container");
    command = NULL;
    asprintf(&command, "mkfs -t %s %s",
        FILESYSTEM_TYPE, config.mapped_device_path);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "creating filesystem failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "creating filesystem inside encrypted container failed");
    }

    // Mount the unlocked container.
    y_log_message(Y_LOG_LEVEL_DEBUG, "mounting new filesystem");
    r = temporary_mount(config.mapped_device_path, DATA_TMP_MOUNTPOINT,
        FILESYSTEM_TYPE);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "mounting encrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "mounting encrypted root device failed");
    }

    // Copy device contents from temporary filesystem to encrypted container.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "copying root device contents from memory");
    command = NULL;
    asprintf(&command, "rsync -a %s/ %s",
        config.tmp_data_location, DATA_TMP_MOUNTPOINT);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "copying from memory to encrypted device failed: return code %d",
            r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying root device contents from memory failed");
    }

    // Unmount filesystem on encrypted data partition.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting encrypted device at %s", DATA_TMP_MOUNTPOINT);
    r = umount(DATA_TMP_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unmounting encrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unmounting encrypted device failed");
    }

    // Lock the container.
    y_log_message(Y_LOG_LEVEL_DEBUG, "locking encrypted device");
    r = encryption_lock(config.mapped_device_name);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "locking encrypted container failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
           "locking container failed");
    }

    // Record that we want to reboot the machine.
    *reboot = true;

    json_decref(json_input);
    r = send_simple_response(response, 200, "status", "ok");
    stop_server();
    return r;
}
