/**
 * Callback function for permanently decrypting an encrypted device.
 * It proceeds as follows:
 * 1.  read the encryption password from the request body;
 * 2.  check if the device isn't already encrypted;
 * 3.  unlock the encrypted container;
 * 4.  mount the filesystem on the encrypted device;
 * 5.  check if there is enough available memory to contain all files currently
 *     stored on the device;
 * 6.  copy device contents to temporary filesystem;
 * 7.  unmount encrypted device;
 * 8.  lock the container;
 * 9.  unmount info partition;
 * 10. remove all partitions from the device;
 * 11. create a new filesystem on the unencrypted device;
 * 12. mount the new filesystem;
 * 13. copy device contents from memory to encrypted container;
 * 14. unmount the new filesystem;
 * 15. reboot.
 * @param[in]   request   incoming HTTP request
 * @param[out]  response  HTTP response to the request
 * @param[in]   user_data extra data to pass between handler and main thread
 * @return                internal status code
 */
int callback_encryption_remove_post(const struct _u_request * request,
    struct _u_response * response, void * user_data)
{
    bool * reboot = (bool *)user_data;
    int r;

    // Read the encryption password from the request body.
    json_t * json_input = ulfius_get_json_body_request(request, NULL);
    const char * password = password = json_string_value(
        json_object_get(json_input, "password"));
    if (password == NULL)
    {
        json_decref(json_input);
        return send_simple_response(response, 400, "error", "missing password");
    }
 
    // Check if the device is encrypted.
    if (! is_encrypted_device(config.data_partition_device))
    {
        // The device is not encrypted, so this command does not make sense.
        json_decref(json_input);
        return send_simple_response(response, 400, "error",
           "not encrypted");
    }

    // Unlock the container.
    y_log_message(Y_LOG_LEVEL_INFO, "unlocking encrypted device");
    r = encryption_unlock(config.data_partition_device, config.mapped_device_name,
        password);
    if (r == -1)
    {
        // The return code -1 signals that the password is wrong.
        json_decref(json_input);
        return send_simple_response(response, 403, "error",
           "incorrect password");
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unlocking encrypted container failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unlocking encrypted container failed");
    }

    // Mount the filesystem on the encrypted data partition.
    r = temporary_mount(config.mapped_device_path, DATA_TMP_MOUNTPOINT,
        FILESYSTEM_TYPE);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "mounting root device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "mounting root device failed");
    }

    // Determine the filesystem usage of the encrypted data partition.
    bool fits = filesystem_fits_in_memory(DATA_TMP_MOUNTPOINT, MEMORY_USAGE);
    if (! fits)
    {
        // Projected memory usage is really high, so abort.
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "device too large");
    }

    // Copy device contents to temporary filesystem.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "copying existing root device contents to memory");
    char * command = NULL;
    asprintf(&command, "rsync -a %s/ %s",
         DATA_TMP_MOUNTPOINT, config.tmp_data_location);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if(r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "copying root device contents into memory failed: return code %d",
            r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying root device contents into memory failed");
    }

    // Unmount encrypted data partition.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting encrypted device at %s", DATA_TMP_MOUNTPOINT);
    r = umount(DATA_TMP_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unmounting encrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unmounting encrypted device failed");
    }

    // Lock the container.
    y_log_message(Y_LOG_LEVEL_DEBUG, "locking encrypted device");
    r = encryption_lock(config.mapped_device_name);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "locking encrypted container failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "locking encrypted container failed");
    }

    // Unmount info partition.
    // We didn't mount this in cryptops-api, but it was mounted by the initrd
    // scripts for use by dropbear.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting info partition at %s", INFO_MOUNTPOINT);
    r = umount(INFO_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unmounting info partition failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
           "unmounting info partition failed");
    }

    // Remove all partitions from the device.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "removing partitions from device %s", config.root_device);
    command = NULL;
    asprintf(&command, "sgdisk -Z %s", config.root_device);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "removing partitions failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "removing partitions failed");
    }

    // Create filesystem on the unencrypted device.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "creating filesystem on unencrypted device");
    command = NULL;
    asprintf(&command, "mkfs -t %s %s", FILESYSTEM_TYPE, config.root_device);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "creating filesystem failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "creating filesystem on unencrypted device failed");
    }

    // Mount the new filesystem.
    y_log_message(Y_LOG_LEVEL_DEBUG, "mounting new filesystem");
    r = temporary_mount(config.root_device, UNENCRYPTED_TMP_MOUNTPOINT,
        FILESYSTEM_TYPE);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "mounting unencrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "mounting unencrypted root device failed");
    }

    // Copy device contents from memory to the unencrypted device.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "copying root device contents from memory");
    command = NULL;
    asprintf(&command, "rsync -a %s/ %s",
        config.tmp_data_location, UNENCRYPTED_TMP_MOUNTPOINT);
    r = system(command);
    if (command)
    {
        free(command);
    }
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "copying from memory to unencrypted device failed: return code %d",
            r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "copying root device contents from memory failed");
    }

    // Unmount filesystem on unencrypted device.
    y_log_message(Y_LOG_LEVEL_DEBUG,
        "unmounting unencrypted device at %s", UNENCRYPTED_TMP_MOUNTPOINT);
    r = umount(UNENCRYPTED_TMP_MOUNTPOINT);
    if (r != 0)
    {
        y_log_message(Y_LOG_LEVEL_ERROR,
            "unmounting unencrypted device failed: return code %d", r);
        json_decref(json_input);
        return send_simple_response(response, 500, "error",
            "unmounting unencrypted device failed");
    }

    // Record that we want to reboot the machine.
    *reboot = true;

    json_decref(json_input);
    r = send_simple_response(response, 200, "status", "ok");
    stop_server();
    return r;
}
